//requires ClanLib (http://clanlib.org); tested only on Linux but theoretically cross-platform
//compile with g++ -o reprap reprap.cpp `pkg-config --cflags --libs clanCore-2.2 clanDisplay-2.2 clanGL-2.2 clanGL1-2.2 clanApp-2.2` -lpthread -ggdb
#include <ClanLib/core.h>
#include <ClanLib/display.h>
#include <ClanLib/gl1.h>
#include <ClanLib/application.h>
#include <cstdlib>
#include <cstdio>
#include <cmath>
//#define M_PI 3.14159265358979323846
class axis_state{
public:
//double Vs,R,kl,km,mu,nu,mnoise;//physical parameters: (voltage,resistance,force/current)/inductance, (force/current,friction,viscous drag,force noise)/mass
//double x,v,I1,I2,Vdir1,Vdir2;//current state: position,velocity,currents,voltage direction setting
double state[10],cov[10][10],noise[10];//order: x,v,I1,I2,Vs,R,kl,km,mu,nu
double x,v,I1,I2,Vs,R,kl,km,mu,nu;//for convenience only
double Vdir1,Vdir2;
int total_size,dynamic_size;
axis_state(double *init_state, double *init_var, double *init_noise,bool actual){
total_size=10;
dynamic_size=4;
for(int i=0;i<total_size;i++){
if(actual){
state[i]=init_state[i];
}
else{
state[i]=init_state[i]+init_noise[i]*(rand()*2.0/RAND_MAX-1);
}
for(int j=0;j<total_size;j++){
cov[i][j]=0;
}
cov[i][i]=init_var[i];
noise[i]=init_noise[i];
}
Vdir1=0;
Vdir2=0;
}
void Euler_step(double t,int num_steps,int mode){
double ts=t/num_steps;
for(int n=0;n<num_steps;n++){
x=state[0];v=state[1];I1=state[2];I2=state[3];Vs=state[4];R=state[5];kl=state[6];km=state[7];mu=state[8];nu=state[9];
double sv,cx=cos((x+ts*v/2)*2*M_PI);
double sx=sin((x+ts*v/2)*2*M_PI);
state[2]=state[2]+ts*(Vs*Vdir1-I1*R-kl*v*cx);
state[3]=state[3]+ts*(Vs*Vdir2-I2*R-kl*v*sx);
if(v<0){
sv=-1;
}
else{
sv=1;
}
state[1]=state[1]+ts*(km*(I1+state[2])/2*cx+km*(I2+state[3])/2*sx-nu*v-sv*mu);
state[0]=state[0]+ts*v;
if(mode==0){//actual
for(int i=0;i<total_size;i++){
state[i]=state[i]+sqrt(ts*noise[i])*(rand()*2.0/RAND_MAX-1);//"wrong" scaling (standard deviation noise[i]/sqrt(3) not noise[i]) is deliberate, as the noise strength will not be known precisely in reality
}
}
if(mode==1){//Kalman
double F[4][10]={{0,1,0,0,0,0,0,0,0,0},{2*M_PI*km*(I2*cx-I1*sx),-nu,km*cx,km*sx,0,0,0,I1*cx+I2*sx,-sv,-v},{2*M_PI*kl*v*sx,-kl*cx,-R,0,Vdir1,-I1,-v*cx,0,0,0},{-2*M_PI*kl*v*cx,-kl*sx,0,-R,Vdir2,-I2,-v*sx,0,0,0}};
for(int i=0;i<dynamic_size;i++){
for(int n=0;n<total_size;n++){
for(int j=0;j<dynamic_size;j++){
cov[i][j]=cov[i][j]+ts*(F[i][n]*cov[n][j]+cov[i][n]*F[j][n]);
}
for(int j=dynamic_size;j<total_size;j++){
cov[i][j]=cov[i][j]+ts*F[i][n]*cov[n][j];
cov[j][i]=cov[j][i]+ts*cov[j][n]*F[i][n];
}}
cov[i][i]=cov[i][i]+ts*noise[i];
}
}
}
x=state[0];v=state[1];I1=state[2];I2=state[3];Vs=state[4];R=state[5];kl=state[6];km=state[7];mu=state[8];nu=state[9];
}
void EKal_adjust(double Imeas,double Ivar,int winding){//Kalman filter: works very well if timestep small enough, but computationally expensive
int c=winding+1;
double K_unscaled[10],Sinv,Idiff;
Sinv=1/(cov[c][c]+Ivar);
Idiff=Imeas-state[c];
for(int i=0;i<total_size;i++){
K_unscaled[i]=cov[i][c];
state[i]=state[i]+K_unscaled[i]*Sinv*Idiff;
}
for(int i=0;i<total_size;i++){
for(int j=0;j<total_size;j++){
cov[i][j]=cov[i][j]-K_unscaled[i]*K_unscaled[j]*Sinv;
}}
x=state[0];v=state[1];I1=state[2];I2=state[3];Vs=state[4];R=state[5];kl=state[6];km=state[7];mu=state[8];nu=state[9];
}
void simple_adjust(double Imeas,int winding){
double Idiff,vdiff,Idv,Idx;
double I_cr=.5,v_cr=10,x_cr=.001,pI_cr=.01,pv_cr=.01;//convergence rate settings for I,v,x,parameters; higher converges faster, but too high is unstable.  These values seem to work but have not been optimized
if(winding==1){
Idiff=Imeas-I1;
Idv=cos(x*2*M_PI);
Idx=-sin(x*2*M_PI);
I1=I1+Idiff*I_cr;
R=R-I1*Idiff*pI_cr;
Vs=Vs+Vdir1*Idiff*pI_cr;
}
else{
Idiff=Imeas-I2;
Idv=sin(x*2*M_PI);
Idx=cos(x*2*M_PI);
I2=I2+Idiff*I_cr;
R=R-I2*Idiff*pI_cr;
Vs=Vs+Vdir2*Idiff*pI_cr;
}
x=x-v*Idx*Idiff*x_cr;
vdiff=-Idv*Idiff*v_cr;
kl=kl-v*Idv*Idiff*pI_cr*1e-4;
km=km+(I1*cos(x*2*M_PI)+I2*sin(x*2*M_PI))*vdiff*pv_cr;
nu=nu-v*vdiff*pv_cr*1e-4;
if(v>0){
mu=mu-vdiff*pv_cr;
}
else{
mu=mu+vdiff*pv_cr;
}
v=v+vdiff;
state[0]=x;state[1]=v;state[2]=I1;state[3]=I2;state[4]=Vs;state[5]=R;state[6]=kl;state[7]=km;state[8]=mu;state[9]=nu;
}
};
int program(const std::vector<CL_String> &args){
CL_SetupCore setup_core;
{
CL_SetupDisplay setup_display;
CL_SetupGL1 target_gl1;
CL_DisplayWindow window("Motor control simulation",1000,700);
CL_GraphicContext gc=window.get_gc();
gc.set_map_mode(cl_map_2d_upper_left);
CL_InputDevice keyboard0=window.get_ic().get_keyboard();
CL_Font font=CL_Font(gc,"FreeSerif",16);
double start_state[10]={0,0,0,0,12/2.8e-3,1.65/2.8e-3,40*5e-4/2.8e-3,40/5e-4,.1/5e-4,1},start_var[10]={.1,.1,.1,.1,100,100,1,1000,100,.5},process_var[10]={.01,1,.01,.01,100,100,.1,100,100,.1};//motor parameters for SY42STH47-1684B (as sold by mendel-parts), 1 unit of x=4 steps=0.5mm,mechanical parameters and noises are wild guesses
axis_state actual(start_state,start_var,process_var,1);
axis_state estimate(start_state,start_var,process_var,0);
double Imax=1.5;
double Inoise=.1;//current measurement noise
double Ivar=Inoise*Inoise;
double tstep=3e-5;//time per calculation cycle; Kalman is unstable above 8e-5 to 10e-5, simple can handle larger
int num_axes=4;//calculation cycles per current reading for this axis: 1 for CPU limited, 4 for A/D converter limited
float last[16];
double xc_p=-2;
double xc_i=-2*tstep;
double xc_d=-.02;
double target_speed=100;//(steps/sec)/4
double target_current,stabilizer_current=0;
double xintegral=0,Imeas;
int n=0;
double t=0;
int plot_x=-1;
bool wait_mode=1;
while(!(keyboard0.get_keycode(CL_KEY_Q))){
n=n+1;
t=t+tstep;
actual.Euler_step(tstep,10,0);
estimate.Euler_step(tstep,1,2);//for Kalman mode, change this 2 to 1, and change both simple_adjust below to EKal_adjust
if(n%(2*num_axes)==0){
Imeas=actual.I1+Inoise*(rand()*2.0/RAND_MAX-1);
if(Imeas*estimate.Vdir1>0){//can't read negative voltages
//estimate.EKal_adjust(Imeas,Ivar,1);
estimate.simple_adjust(Imeas,1);
}
}
if(n%(2*num_axes)==num_axes){
Imeas=actual.I2+Inoise*(rand()*2.0/RAND_MAX-1);
if(Imeas*estimate.Vdir2>0){
//estimate.EKal_adjust(Imeas,Ivar,2);
estimate.simple_adjust(Imeas,2);
}
}
xintegral=xintegral+estimate.x-target_speed*t;
target_current=xc_p*(estimate.x-target_speed*t)+xc_d*(estimate.v-target_speed)+xc_i*xintegral;
if(target_current>Imax){
target_current=Imax;
}
if(target_current<-Imax){
target_current=-Imax;
}
if(estimate.I1<target_current*cos(estimate.x*2*M_PI)+stabilizer_current*sin(estimate.x*2*M_PI)){
actual.Vdir1=1;
estimate.Vdir1=1;
}
else{
actual.Vdir1=-1;
estimate.Vdir1=-1;
}
if(estimate.I2<target_current*sin(estimate.x*2*M_PI)-stabilizer_current*cos(estimate.x*2*M_PI)){
actual.Vdir2=1;
estimate.Vdir2=1;
}
else{
actual.Vdir2=-1;
estimate.Vdir2=-1;
}
if(plot_x>0){
CL_Draw::line(gc,plot_x-1,last[0],plot_x,350-200*actual.I1,CL_Colorf::red);
last[0]=350-200*actual.I1;
CL_Draw::line(gc,plot_x-1,last[1],plot_x,350-200*actual.I2,CL_Colorf::green);
last[1]=350-200*actual.I2;
CL_Draw::line(gc,plot_x-1,last[2],plot_x,350-200*estimate.I1,CL_Colorf::darkred);
last[2]=350-200*estimate.I1;
CL_Draw::line(gc,plot_x-1,last[3],plot_x,350-200*estimate.I2,CL_Colorf::darkgreen);
last[3]=350-200*estimate.I2;
CL_Draw::line(gc,plot_x-1,last[4],plot_x,700*fmod(1-fmod(actual.x,1),1),CL_Colorf::blue);
last[4]=700*fmod(1-fmod(actual.x,1),1);
CL_Draw::line(gc,plot_x-1,last[5],plot_x,700*fmod(1-fmod(estimate.x,1),1),CL_Colorf::darkblue);
last[5]=700*fmod(1-fmod(estimate.x,1),1);
CL_Draw::line(gc,plot_x-1,last[6],plot_x,350-3*actual.v,CL_Colorf(1.0f,0,1));
last[6]=350-3*actual.v;
CL_Draw::line(gc,plot_x-1,last[7],plot_x,350-3*estimate.v,CL_Colorf(.5f,0,.5));
last[7]=350-3*estimate.v;
plot_x=plot_x+1;
if(plot_x>1000){
plot_x=-1;
window.flip();
while(wait_mode && !(keyboard0.get_keycode(CL_KEY_SPACE)) && !(keyboard0.get_keycode(CL_KEY_Q)) && !(keyboard0.get_keycode(CL_KEY_R))){}
if(keyboard0.get_keycode(CL_KEY_R)){wait_mode=0;}
}
}
CL_KeepAlive::process();
if((n==1)||(keyboard0.get_keycode(CL_KEY_SPACE))){
gc.clear(CL_Colorf::white);
CL_Draw::line(gc,0,350,1000,350,CL_Colorf::yellow);
plot_x=1;
font.draw_text(gc,10,10,"t="+CL_StringHelp::double_to_text(t)+"sec Actual x="+CL_StringHelp::double_to_text(actual.x)+"Estimate x="+CL_StringHelp::double_to_text(estimate.x)+" Press Space to draw another plot, R to run without continuously plotting, or Q to quit",CL_Colorf::black);
}
}
}
return 0;
}
CL_ClanApplication app(&program);
